import datetime

import pandas as pd
import numpy as np

import os
import orekit
orekit.initVM()
import java.io
import org.orekit

DM = org.orekit.data.DataContext.getDefault().getDataProvidersManager()
DM.clearProviders()
orekit_data_folder = java.io.File(os.getenv('OREKIT_DATA'))  # Replace by your orekit data folder
DM.addProvider(org.orekit.data.DirectoryCrawler(orekit_data_folder))

# Data from the NeQuickG reference document
# (https://www.gsc-europa.eu/sites/default/files/sites/all/files/Galileo_Ionospheric_Model.pdf)

TEST_DATA = {
    "high_solar_activity": {
        "test_file": "data/high_solar_activity.csv",
        "alpha": [236.831641, -0.39362878, 0.00402826613],
    },
    "medium_solar_activity": {
        "test_file": "data/medium_solar_activity.csv",
        "alpha": [121.129893, 0.351254133, 0.0134635348],
    },
    "low_solar_activity": {
        "test_file": "data/low_solar_activity.csv",
        "alpha": [2.580271, 0.127628236, 0.0252748384],
    }
}


def read_test_data(filename: str) -> pd.DataFrame:
    df = pd.read_csv(filename, sep=";")
    df['Station latitude'] = np.radians(df['Station latitude'])
    df['Station longitude'] = np.radians(df['Station longitude'])
    df['Satellite latitude'] = np.radians(df['Satellite latitude'])
    df['Satellite longitude'] = np.radians(df['Satellite longitude'])

    months = df['Month'].values
    times = df['UTC Time'].values
    utc_datetimes = [datetime.datetime(year=2024, month=months[i], day=1, hour=times[i]) for i in range(len(months))]

    df['UTC Datetime'] = utc_datetimes

    return df


def nequick_stec(rcv_latlonalt, sat_latlonalt, utc_datetime, alpha):
    rcv_latlonalt = np.array(rcv_latlonalt).tolist()
    sat_latlonalt = np.array(sat_latlonalt).tolist()
    alpha = np.array(alpha).tolist()
    nequick_model = org.orekit.models.earth.ionosphere.NeQuickModel(alpha)

    utc_timescale = org.orekit.time.TimeScalesFactory.getUTC()
    absolute_date = org.orekit.time.AbsoluteDate(utc_datetime.isoformat(), utc_timescale)
    rec_geo = org.orekit.bodies.GeodeticPoint(*rcv_latlonalt)
    sat_geo = org.orekit.bodies.GeodeticPoint(*sat_latlonalt)
    stec = nequick_model.stec(absolute_date, rec_geo, sat_geo)

    return stec


def test_csv(test_case: str):
    print('Reading test data... ', end='')
    df = read_test_data(TEST_DATA[test_case]["test_file"])
    print('Done.')

    alpha = TEST_DATA[test_case]['alpha']
    errors_tecu = []
    errors_meters = []

    freq_L1 = 1575.42 * 10 ** 6
    freq_coeff = (40.32 / (freq_L1 ** 2) * 10 ** 16)


    print('=====================================')
    print(f'Running test {test_case}')
    for _, row in df.iterrows():
        t = row['UTC Datetime']

        sat_lat = row['Satellite latitude']
        sat_lon = row['Satellite longitude']
        sat_height = row['Satellite height']
        sat_pos = [sat_lat, sat_lon, sat_height]

        rcv_lat = row['Station latitude']
        rcv_lon = row['Station longitude']
        rcv_height = row['Station height']
        rcv_pos = [rcv_lat, rcv_lon, rcv_height]

        expected_stec = row['STEC (TECU)']
        orekit_stec = nequick_stec(rcv_pos, sat_pos, t, alpha)

        errors_tecu.append(expected_stec - orekit_stec)
        errors_meters.append((expected_stec - orekit_stec) * freq_coeff)

        print(f'\nTest at time {t}')
        print(f'\t- Satellite position: {sat_pos}')
        print(f'\t- Station position: {rcv_pos}')
        print(f'\t- Expected STEC: {expected_stec} TECU')
        print(f'\t- Computed STEC: {orekit_stec} TECU')
        print(f'\t- Difference: {expected_stec - orekit_stec} TECU')
        print(f'\t- Difference: {(expected_stec - orekit_stec) * freq_coeff} meters of L1 delay')

    print('\nDone.')
    print(f'\t- Max difference: {np.max(np.array(errors_meters))} meters of L1 delay')
    print(f'\t- Max difference: {np.max(np.array(errors_tecu))} TECU')
    print(f'\t- Mean difference: {np.sqrt(np.square(errors_meters).mean())} meters of L1 delay')
    print(f'\t- Mean difference: {np.sqrt(np.square(errors_tecu).mean())} TECU')

    return errors_meters, errors_tecu


def test_orekit():
    # List of tests already implemented into orekit
    high_alpha = [236.831641, -0.39362878, 0.00402826613]
    medium_alpha = [121.129893, 0.351254133, 0.0134635348]

    freq_L1 = 1575.42 * 10 ** 6
    freq_coeff = (40.32 / (freq_L1 ** 2) * 10 ** 16)

    tests = [
        [[np.radians(82.49), np.radians(297.66), 78.11], [np.radians(54.29), np.radians(8.23), 20281546.18], datetime.datetime(year=2018, month=4, day=2, hour=0), high_alpha, 20.40],
        [[np.radians(-31.80), np.radians(115.89), 12.78], [np.radians(-14.31), np.radians(124.09), 20100697.90], datetime.datetime(year=2018, month=4, day=2, hour=16), medium_alpha, 6.96],
    ]

    print('==============================')
    print('Running orekit tests')
    for rcv_pos, sat_pos, t, alpha, expected_stec in tests:
        orekit_stec = nequick_stec(rcv_pos, sat_pos, t, alpha)

        print(f'\nTest at time {t}')
        print(f'\t- Satellite position: {sat_pos}')
        print(f'\t- Station position: {rcv_pos}')
        print(f'\t- Expected STEC: {expected_stec} TECU')
        print(f'\t- Computed STEC: {orekit_stec} TECU')
        print(f'\t- Difference: {expected_stec - orekit_stec} TECU')
        print(f'\t- Difference: {(expected_stec - orekit_stec) * freq_coeff} meters of L1 delay')


if __name__ == "__main__":
    all_meters, all_tecu = [], []
    for test_case in TEST_DATA:
        errors_meters, errors_tecu = test_csv(test_case)
        all_meters += errors_meters
        all_tecu += errors_tecu

    print('===============')
    print('Overall stats')
    print(f'\t- Max difference: {np.max(np.array(all_meters))} meters of L1 delay')
    print(f'\t- Max difference: {np.max(np.array(all_tecu))} TECU')
    print(f'\t- Mean difference: {np.sqrt(np.square(all_meters).mean())} meters of L1 delay')
    print(f'\t- Mean difference: {np.sqrt(np.square(all_tecu).mean())} TECU')

    test_orekit()
